import 'dart:async';
import 'dart:collection';
import 'dart:developer';
import 'dart:io';

import 'package:events2/events2.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_foreground_task/flutter_foreground_task.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import '../../../utils/constants.dart';

import '../../../rtc.dart';
import '../../../videosdk.dart';
import '../../videosdk_client/videosdk_client.dart';
import '../me/bloc/me_bloc.dart';
import '../media_devices/bloc/media_devices_bloc.dart';
import '../peers/bloc/peers_bloc.dart';
import '../producers/bloc/producers_bloc.dart';
import '../room/bloc/room_bloc.dart';
import 'web_socket.dart';

const VIDEO_LAYERS = {
  "sd": {
    "low": {"s": 0, 't': 2},
    "med": {"s": 0, 't': 2},
    "high": {"s": 1, 't': 2},
    "s0t0": {"s": 0, 't': 0},
    "s0t1": {"s": 0, 't': 1},
    "s0t2": {"s": 0, 't': 2},
    "s1t0": {"s": 1, 't': 0},
    "s1t1": {"s": 1, 't': 1},
    "s1t2": {"s": 1, 't': 2},
    "s2t0": {"s": 2, 't': 0},
    "s2t1": {"s": 2, 't': 1},
    "s2t2": {"s": 2, 't': 2},
  },
  "hd": {
    "low": {"s": 0, "t": 2},
    "med": {"s": 1, "t": 2},
    "high": {"s": 2, "t": 2},
    "s0t0": {"s": 0, "t": 0},
    "s0t1": {"s": 0, "t": 1},
    "s0t2": {"s": 0, "t": 2},
    "s1t0": {"s": 1, "t": 0},
    "s1t1": {"s": 1, "t": 1},
    "s1t2": {"s": 1, "t": 2},
    "s2t0": {"s": 2, "t": 0},
    "s2t1": {"s": 2, "t": 1},
    "s2t2": {"s": 2, "t": 2},
  },
};

class RoomClientRepository {
  final ProducersBloc producersBloc;
  final PeersBloc peersBloc;
  final MeBloc meBloc;
  final RoomBloc roomBloc;
  final MediaDevicesBloc mediaDevicesBloc;

  final String token;
  final String meetingId;
  final String peerId;
  late NotificationInfo notification;

  final String displayName;
  final bool micEnabled;
  final bool webcamEnabled;
  late final EventEmitter eventEmitter;
  String maxResolution;

  bool _closed = false;

  WebSocket? _webSocket;
  Device? _mediasoupDevice;
  Transport? _sendTransport;
  Transport? _recvTransport;
  bool _produce = false;
  bool _consume = true;
  StreamSubscription<MediaDevicesState>? _mediaDevicesBlocSubscription;
  String? audioInputDeviceId;
  String? audioOutputDeviceId;
  String? videoInputDeviceId;

  var _usageIntervalDuration = 1;

  var _usageDataPointsPeerMaxRes = Map();
  late Timer _usageInterval;

  //audio consumption interval
  var _audioConsumption = 0;

  Function(RoomClientRepository) onMeetingJoined;

  RoomClientRepository({
    required this.token,
    required this.micEnabled,
    required this.webcamEnabled,
    required this.meetingId,
    required this.peerId,
    required this.displayName,
    required this.maxResolution,
    required this.notification,
    // blocs
    required this.producersBloc,
    required this.peersBloc,
    required this.meBloc,
    required this.roomBloc,
    required this.mediaDevicesBloc,
    //
    required this.onMeetingJoined,
    required this.eventEmitter,
  }) {
    _mediaDevicesBlocSubscription =
        mediaDevicesBloc.stream.listen((MediaDevicesState state) async {
      if (state.selectedAudioInput != null &&
              state.selectedAudioInput?.deviceId != audioInputDeviceId
          // &&
          // this.micEnabled
          ) {
        await disableMic();
        enableMic();
      }

      if (state.selectedVideoInput != null &&
              state.selectedVideoInput?.deviceId != videoInputDeviceId
          // &&
          // this.webcamEnabled
          ) {
        await disableWebcam();
        enableWebcam(true);
      }
    });
  }

  void close() {
    if (_closed) {
      return;
    }

    _webSocket?.close();
    _sendTransport?.close();
    _recvTransport?.close();
    _mediaDevicesBlocSubscription?.cancel();

    if (this._usageInterval.isActive) this._usageInterval.cancel();
    this.eventEmitter.emit("meeting-left");
  }

  Future<void> disableMic() async {
    if (producersBloc.state.mic != null) {
      String micId = producersBloc.state.mic!.id;

      producersBloc.add(ProducerRemove(source: 'mic'));

      try {
        await _webSocket!.socket.request('closeProducer', {
          'producerId': micId,
        });
      } catch (error) {}
    }
  }

  Future<void> disableWebcam() async {
    if (producersBloc.state.webcam?.id != null) {
      meBloc.add(MeSetWebcamInProgress(progress: true));
      String webcamId = producersBloc.state.webcam!.id;

      producersBloc.add(ProducerRemove(source: 'webcam'));

      try {
        await _webSocket!.socket.request('closeProducer', {
          'producerId': webcamId,
        });
      } catch (error) {
      } finally {
        meBloc.add(MeSetWebcamInProgress(progress: false));
      }
    }
  }

  Future<void> disableShare() async {
    if (producersBloc.state.screen?.id != null) {
      meBloc.add(MeSetScreenshareInProgress(progress: true));
      String screenId = producersBloc.state.screen!.id;

      producersBloc.add(ProducerRemove(source: "screen"));

      try {
        await _webSocket!.socket.request('closeProducer', {
          'producerId': screenId,
        });
      } catch (error) {
        log("disableShare error => $error");
      } finally {
        FlutterForegroundTask.stopService();

        meBloc.add(MeSetScreenshareInProgress(progress: false));
      }
    }
  }

  Future<void> pauseConsumer(Consumer consumer) async {
    if (consumer.paused) return;
    try {
      await _webSocket!.socket.request('pauseConsumer', {
        'consumerId': consumer.id,
      });

      peersBloc.add(PeerPausedConsumer(consumerId: consumer.id));
    } catch (error) {}
  }

  Future<void> resumeConsumer(Consumer consumer) async {
    if (!consumer.paused) return;
    try {
      await _webSocket!.socket.request('resumeConsumer', {
        'consumerId': consumer.id,
      });

      peersBloc.add(PeerResumedConsumer(consumerId: consumer.id));
    } catch (error) {}
  }

  Future<void> muteMic() async {
    producersBloc.add(ProducerPaused(source: 'mic'));

    try {
      await _webSocket!.socket.request('pauseProducer', {
        'producerId': producersBloc.state.mic!.id,
      });
    } catch (error) {}
  }

  Future<void> unmuteMic() async {
    producersBloc.add(ProducerResumed(source: 'mic'));

    try {
      await _webSocket!.socket.request('resumeProducer', {
        'producerId': producersBloc.state.mic!.id,
      });
    } catch (error) {}
  }

  Future<void> startRecording(String webhookUrl) async {
    try {
      await _webSocket!.socket.request('startRecording', {
        'webhookUrl': webhookUrl,
      });
    } catch (error) {}
  }

  Future<void> stopRecording() async {
    try {
      await _webSocket!.socket.request('stopRecording', {});
    } catch (error) {}
  }

  Future<void> startLivestream(outputs) async {
    try {
      await _webSocket!.socket.request('startLivestream', {"outputs": outputs});
    } catch (error) {}
  }

  Future<void> stopLivestream() async {
    try {
      await _webSocket!.socket.request('stopLivestream', {});
    } catch (error) {}
  }

  void changeWebcam(String deviceId) async {
    if (mediaDevicesBloc.state.selectedVideoInput!.deviceId == deviceId) {
      return;
    }

    try {
      MediaDeviceInfo device = mediaDevicesBloc.state.videoInputs
          .firstWhere((element) => element.deviceId == deviceId);

      mediaDevicesBloc.add(MediaDeviceSelectVideoInput(device));

      // await disableWebcam();
      // enableWebcam();
    } catch (err) {
      log("handle webcam error $err");
    }
  }

  List<MediaDeviceInfo> getWebcams() {
    return mediaDevicesBloc.state.videoInputs;
  }

  String? get selectedWebcamId {
    return mediaDevicesBloc.state.selectedVideoInput!.deviceId;
  }

  List<MediaDeviceInfo> getMics() {
    return mediaDevicesBloc.state.audioInputs;
  }

  String? get selectedMicId {
    return mediaDevicesBloc.state.selectedAudioInput!.deviceId;
  }

  Future<void> enablePeerMic(String peerId) async {
    try {
      await _webSocket!.socket.request('enablePeerMic', {
        "peerId": peerId,
      });
    } catch (error) {}
  }

  Future<void> disablePeerMic(String peerId) async {
    try {
      await _webSocket!.socket.request('disablePeerMic', {
        "peerId": peerId,
      });
    } catch (error) {}
  }

  Future<void> enablePeerWebcam(String peerId) async {
    try {
      await _webSocket!.socket.request('enablePeerWebcam', {
        "peerId": peerId,
      });
    } catch (error) {}
  }

  Future<void> disablePeerWebcam(String peerId) async {
    try {
      await _webSocket!.socket.request('disablePeerWebcam', {
        "peerId": peerId,
      });
    } catch (error) {}
  }

  Future<void> removePeer(String peerId) async {
    try {
      await _webSocket!.socket.request('removePeer', {
        "peerId": peerId,
      });
    } catch (error) {}
  }

  void _producerCallback(Producer producer) {
    if (producer.source == 'mic') {
      producer.on('transportclose', () {
        producersBloc.add(ProducerRemove(source: 'mic'));
      });

      producer.on('trackended', () {
        disableMic().catchError((data) {});
      });
    } else if (producer.source == 'webcam') {
      producer.on('transportclose', () {
        producersBloc.add(ProducerRemove(source: 'webcam'));
      });

      producer.on('trackended', () {
        disableWebcam().catchError((data) {});
      });

      meBloc.add(MeSetWebcamInProgress(progress: false));
    } else if (producer.source == 'share') {
      producer.on('transportclose', () {
        FlutterForegroundTask.stopService();
        producersBloc.add(ProducerRemove(source: 'share'));
      });

      producer.on('trackended', () {
        FlutterForegroundTask.stopService();
        disableShare().catchError((data) {});
      });

      meBloc.add(MeSetScreenshareInProgress(progress: false));
    }

    producersBloc.add(ProducerAdd(producer: producer));
  }

  void _consumerCallback(Consumer consumer, [dynamic accept]) {
    // consumer.on('transportclose', () {
    //   // consumersBloc.add(ConsumerRemove(consumerId: consumer.id));
    // });

    ScalabilityMode scalabilityMode = ScalabilityMode.parse(
      consumer.rtpParameters.encodings.first.scalabilityMode,
    );

    accept({});

    peersBloc.add(PeerAddConsumer(peerId: consumer.peerId, consumer: consumer));
  }

  Future<MediaStream> createAudioStream() async {
    audioInputDeviceId = mediaDevicesBloc.state.selectedAudioInput!.deviceId;
    Map<String, dynamic> mediaConstraints = {
      'audio': {
        'optional': [
          {
            'sourceId': audioInputDeviceId,
          },
        ],
      },
    };

    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    return stream;
  }

  Future<MediaStream> createVideoStream() async {
    videoInputDeviceId = mediaDevicesBloc.state.selectedVideoInput?.deviceId;

    Map<String, dynamic> mediaConstraints = <String, dynamic>{
      'audio': false,
      'video': {
        'mandatory': {
          'minWidth':
              '1280', // Provide your own width, height and frame rate here
          'minHeight': '720',
          'minFrameRate': '30',
        },
        'optional': [
          {
            'sourceId': videoInputDeviceId,
          },
        ],
      },
    };
    MediaStream stream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);

    return stream;
  }

  Future<MediaStream> createShareStream() async {
    final mediaConstraints = <String, dynamic>{'audio': true, 'video': true};

    MediaStream stream =
        await navigator.mediaDevices.getDisplayMedia(mediaConstraints);

    return stream;
  }

  void enableWebcam([bool dontCheckProgress = false]) async {
    if (meBloc.state.webcamInProgress && !dontCheckProgress) {
      return;
    }

    meBloc.add(MeSetWebcamInProgress(progress: true));

    if (_mediasoupDevice?.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) ==
        false) {
      return;
    }

    MediaStream? videoStream;
    MediaStreamTrack? track;

    try {
      // NOTE: prefer using h264
      RtpCodecCapability? codec = _mediasoupDevice!.rtpCapabilities.codecs
          .firstWhere((RtpCodecCapability c) {
        return c.mimeType.toLowerCase() == 'video/vp9' ||
            c.mimeType.toLowerCase() == 'video/vp8';
      },
              // (RtpCodecCapability c) => c.mimeType.toLowerCase() == 'video/h264',
              orElse: () =>
                  throw 'desired vp9 codec+configuration is not supported');

      videoStream = await createVideoStream();

      track = videoStream.getVideoTracks().first;

      meBloc.add(MeSetWebcamInProgress(progress: true));

      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(
          videoGoogleStartBitrate: 1000,
        ),
        encodings: kIsWeb
            ? [
                RtpEncodingParameters(
                    scalabilityMode: 'S3T3_KEY', scaleResolutionDownBy: 1.0),
              ]
            : [],
        stream: videoStream,
        appData: {'source': 'webcam'},
        source: 'webcam',
        codec: codec,
      );
    } catch (error) {
      meBloc.add(MeSetWebcamInProgress(progress: false));

      if (videoStream != null) {
        await videoStream.dispose();
      }
    }
  }

  void enableMic() async {
    if (_mediasoupDevice?.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) ==
        false) {
      return;
    }

    MediaStream? audioStream;
    MediaStreamTrack? track;
    try {
      audioStream = await createAudioStream();
      track = audioStream.getAudioTracks().first;
      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(opusStereo: 1, opusDtx: 1),
        stream: audioStream,
        appData: {
          'source': 'mic',
        },
        source: 'mic',
      );
    } catch (error) {
      if (audioStream != null) {
        await audioStream.dispose();
      }
    }
  }

  void enableShare() async {
    if (meBloc.state.shareInProgress) {
      return;
    }

    try {
      await FlutterForegroundTask.startService(
        notificationTitle: this.notification.title,
        notificationText: this.notification.message,
      );
    } catch (error) {
      log("error => $error");
    }

    meBloc.add(MeSetScreenshareInProgress(progress: true));

    MediaStream? shareStream;
    MediaStreamTrack? track;

    try {
      RtpCodecCapability? codec = _mediasoupDevice!.rtpCapabilities.codecs
          .firstWhere((RtpCodecCapability c) {
        return c.mimeType.toLowerCase() == 'video/vp9' ||
            c.mimeType.toLowerCase() == 'video/vp8';
      },
              // (RtpCodecCapability c) => c.mimeType.toLowerCase() == 'video/h264',
              orElse: () =>
                  throw 'desired vp9 codec+configuration is not supported');

      shareStream = await createShareStream();

      track = shareStream.getVideoTracks().first;

      meBloc.add(MeSetScreenshareInProgress(progress: true));

      _sendTransport!.produce(
        track: track,
        codecOptions: ProducerCodecOptions(
          videoGoogleStartBitrate: 1000,
        ),
        encodings: kIsWeb
            ? [
                RtpEncodingParameters(
                    scalabilityMode: 'S3T3_KEY', scaleResolutionDownBy: 1.0),
              ]
            : [],
        stream: shareStream,
        appData: {'share': true},
        source: 'screen',
        codec: codec,
      );
    } catch (error) {
      log('error enable share  error => $error');

      meBloc.add(MeSetScreenshareInProgress(progress: false));

      FlutterForegroundTask.stopService();

      if (shareStream != null) {
        await shareStream.dispose();
      }
    }
  }

  void setConsumerQuality(consumerId, String quality) async {
    late int? spatialLayers;
    late int? temporalLayers;

    try {
      if (quality == 'low' || quality == 'med' || quality == 'high') {
        final layers = VIDEO_LAYERS[this.maxResolution]![quality];
        spatialLayers = layers?['s'];
        temporalLayers = layers?['t'];
      } else {
        String sCh = quality[0];
        String tCh = quality[2];
        int s = int.parse(quality[1]);
        int t = int.parse(quality[3]);

        if (s > 2 || t > 2 || sCh != "s" || tCh != "t") {
          throw ("");
        }

        spatialLayers = s;
        temporalLayers = t;
      }

      setConsumerPreferredLayers(consumerId, spatialLayers, temporalLayers);
    } catch (error) {
      log("invalid quality");
    }
  }

  void setConsumerPreferredLayers(
    consumerId,
    spatialLayer,
    temporalLayer,
  ) async {
    try {
      await _webSocket!.socket.request('setConsumerPreferredLayers', {
        "consumerId": consumerId,
        "spatialLayer": spatialLayer,
        "temporalLayer": temporalLayer,
      });
    } catch (error) {}
  }

  Future<void> _joinRoom() async {
    try {
      _mediasoupDevice = Device();

      dynamic routerRtpCapabilities =
          await _webSocket!.socket.request('getRouterRtpCapabilities', {});

      final rtpCapabilities = RtpCapabilities.fromMap(routerRtpCapabilities);
      rtpCapabilities.headerExtensions
          .removeWhere((he) => he.uri == 'urn:3gpp:video-orientation');
      await _mediasoupDevice!.load(routerRtpCapabilities: rtpCapabilities);

      if (_mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeAudio) ==
              true ||
          _mediasoupDevice!.canProduce(RTCRtpMediaType.RTCRtpMediaTypeVideo) ==
              true) {
        _produce = true;
      }

      if (_produce) {
        Map transportInfo =
            await _webSocket!.socket.request('createWebRtcTransport', {
          'forceTcp': false,
          'producing': true,
          'consuming': false,
          'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
        });

        _sendTransport = _mediasoupDevice!.createSendTransportFromMap(
          transportInfo,
          producerCallback: _producerCallback,
        );

        _sendTransport!.on('connect', (Map data) {
          _webSocket!.socket
              .request('connectWebRtcTransport', {
                'transportId': _sendTransport!.id,
                'dtlsParameters': data['dtlsParameters'].toMap(),
              })
              .then(data['callback'])
              .catchError(data['errback']);
        });

        _sendTransport!.on('produce', (Map data) async {
          try {
            Map response = await _webSocket!.socket.request(
              'produce',
              {
                'transportId': _sendTransport!.id,
                'kind': data['kind'],
                'rtpParameters': data['rtpParameters'].toMap(),
                if (data['appData'] != null)
                  'appData': Map<String, dynamic>.from(data['appData'])
              },
            );

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });

        _sendTransport!.on('producedata', (data) async {
          try {
            Map response = await _webSocket!.socket.request('produceData', {
              'transportId': _sendTransport!.id,
              'sctpStreamParameters': data['sctpStreamParameters'].toMap(),
              'label': data['label'],
              'protocol': data['protocol'],
              'appData': data['appData'],
            });

            data['callback'](response['id']);
          } catch (error) {
            data['errback'](error);
          }
        });
      }

      if (_consume) {
        Map transportInfo = await _webSocket!.socket.request(
          'createWebRtcTransport',
          {
            'forceTcp': false,
            'producing': false,
            'consuming': true,
            'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
          },
        );

        _recvTransport = _mediasoupDevice!.createRecvTransportFromMap(
          transportInfo,
          consumerCallback: _consumerCallback,
        );

        _recvTransport!.on(
          'connect',
          (data) {
            _webSocket!.socket
                .request(
                  'connectWebRtcTransport',
                  {
                    'transportId': _recvTransport!.id,
                    'dtlsParameters': data['dtlsParameters'].toMap(),
                  },
                )
                .then(data['callback'])
                .catchError(data['errback']);
          },
        );
      }

      final deviceInfo = await getDeviceInfo();

      Map response = await _webSocket!.socket.request('join', {
        'displayName': displayName,
        'device': {
          'name': "Flutter",
          'flag': 'flutter',
          'version': sdkVersion,
        },
        'rtpCapabilities': _mediasoupDevice!.rtpCapabilities.toMap(),
        'sctpCapabilities': _mediasoupDevice!.sctpCapabilities.toMap(),
        'deviceInfo': deviceInfo,
      });

      response['peers'].forEach((value) {
        peersBloc.add(PeerAdd(newPeer: value));
      });

      if (_produce) {
        if (micEnabled) {
          enableMic();
        }
        if (webcamEnabled) {
          enableWebcam();
        }

        _sendTransport!.on('connectionstatechange', (connectionState) {
          if (connectionState == 'connected') {
            // enableChatDataProducer();
            // enableBotDataProducer();
          }
        });
      }

      this._usageInterval = Timer.periodic(
        Duration(seconds: _usageIntervalDuration),
        (timer) => this._updateUsageStats(),
      );

      this.eventEmitter.emit('meeting-joined');

      await FlutterForegroundTask.init(
        androidNotificationOptions: AndroidNotificationOptions(
          channelId: 'notification_share_screen',
          channelName: 'Foreground Notification',
          channelDescription:
              'This notification appears when the foreground service is running.',
          channelImportance: NotificationChannelImportance.LOW,
          priority: NotificationPriority.LOW,
          iconData: NotificationIconData(
            resType: ResourceType.drawable,
            resPrefix: ResourcePrefix.ic,
            name: this.notification.icon,
          ),
        ),
        iosNotificationOptions: const IOSNotificationOptions(
          showNotification: true,
          playSound: false,
        ),
        foregroundTaskOptions: const ForegroundTaskOptions(
          interval: 5000,
          autoRunOnBoot: true,
          allowWifiLock: true,
        ),
        printDevLog: true,
      );
    } catch (error) {
      log(error.toString());
      close();
    }
  }

  void join() {
    _webSocket = WebSocket(peerId: peerId, meetingId: meetingId, token: token);

    _webSocket!.onOpen = requestEntry;
    _webSocket!.onFail = () {};
    _webSocket!.onDisconnected = () {
      if (_sendTransport != null) {
        _sendTransport!.close();
        _sendTransport = null;
      }
      if (_recvTransport != null) {
        _recvTransport!.close();
        _recvTransport = null;
      }
    };
    _webSocket!.onClose = () {
      if (_closed) return;

      close();
    };

    _webSocket!.onRequest = (request, accept, reject) async {
      switch (request['method']) {
        case "close":
          {
            this.close();
            break;
          }

        case 'newConsumer':
          {
            if (!_consume) {
              reject(403, 'I do not want to consume');
              break;
            }
            try {
              _recvTransport!.consume(
                id: request['data']['id'],
                producerId: request['data']['producerId'],
                kind: RTCRtpMediaTypeExtension.fromString(
                    request['data']['kind']),
                rtpParameters:
                    RtpParameters.fromMap(request['data']['rtpParameters']),
                appData: Map<String, dynamic>.from(request['data']['appData']),
                peerId: request['data']['peerId'],
                accept: accept,
              );
            } catch (error) {
              throw (error);
            }
            break;
          }

        case 'enableMic':
          {
            try {
              this.eventEmitter.emit('mic-requested', {
                // 'participantId': peerId,
                "accept": () {
                  this.enableMic();
                },
                "reject": () {},
              });

              accept();
            } catch (error) {
              throw error;
            }
            break;
          }
        case 'disableMic':
          {
            try {
              this.disableMic();
              accept();
            } catch (error) {
              throw error;
            }
            break;
          }
        case 'enableWebcam':
          {
            try {
              this.eventEmitter.emit('webcam-requested', {
                // "participantId": peerId,
                "accept": () {
                  this.enableWebcam();
                },
                "reject": () {},
              });

              accept();
            } catch (error) {
              throw error;
            }
            break;
          }
        case 'disableWebcam':
          {
            try {
              this.disableWebcam();
              accept();
            } catch (error) {
              throw error;
            }
            break;
          }
        case 'dataUsage':
          {
            //final document to send
            var consumptionDoc = <String, dynamic>{};

            //? summation for this all consumers streams.
            num resolutionTotal = 0;

            this._usageDataPointsPeerMaxRes.forEach((peerData, maxRes) {
              resolutionTotal += maxRes as num;

              //if maxRes not exists then set first
              if (consumptionDoc[maxRes.toString()] == null) {
                consumptionDoc[maxRes.toString()] = [];
              }

              consumptionDoc[maxRes.toString()].add(peerData);
            });

            //get rounded resolution
            resolutionTotal = this._getRoundedResolution(resolutionTotal);

            var data = <String, dynamic>{
              "data": <String, dynamic>{
                "peerId": this.peerId,
                "consumption": consumptionDoc,
                "audioConsumption": this._audioConsumption,
              },
              "totalResolution": resolutionTotal,
            };

            accept(data);

            //flush the data points after sending data
            this._audioConsumption = 0;
            this._usageDataPointsPeerMaxRes = Map();
            break;
          }
        default:
          break;
      }
    };

    _webSocket!.onNotification = (notification) async {
      switch (notification['method']) {
        case 'producerScore':
          {
            break;
          }

        case 'consumerClosed':
          {
            String consumerId = notification['data']['consumerId'];
            peersBloc.add(PeerRemoveConsumer(consumerId: consumerId));

            break;
          }

        case 'consumerPaused':
          {
            // String consumerId = notification['data']['consumerId'];
            // peersBloc.add(PeerPausedConsumer(consumerId: consumerId));
            break;
          }

        case 'consumerResumed':
          {
            // String consumerId = notification['data']['consumerId'];
            // peersBloc.add(PeerResumedConsumer(consumerId: consumerId));
            break;
          }

        case "consumerLayersChanged":
          {
            String consumerId = notification['data']['consumerId'];
            int spatialLayer = notification['data']['spatialLayer'] ?? -1;
            int temporalLayer = notification['data']['temporalLayer'] ?? -1;

            peersBloc.add(PeerChangeConsumerQuality(
              consumerId: consumerId,
              spatialLayer: spatialLayer,
              temporalLayer: temporalLayer,
            ));

            Consumer? consumer = _recvTransport?.getConsumer(consumerId);

            if (consumer != null) {
              consumer.spatialLayer = spatialLayer;
              consumer.temporalLayer = temporalLayer;
            }

            break;
          }

        case "entryRequested":
          {
            var id = notification['data']['id'];
            var name = notification['data']['id'];

            if (id != this.peerId) {
              this.eventEmitter.emit(
                  "ENTRY_REQUESTED", <String, dynamic>{'id': id, 'name': name});
            }

            break;
          }

        case "entryResponded":
          {
            var id = notification['data']['id'];
            var decision = notification['data']['decision'];

            this.eventEmitter.emit("ENTRY_RESPONDED",
                <String, dynamic>{'id': id, 'decision': decision});

            if (id == this.peerId && decision == "allowed") {
              this._joinRoom();
            }

            break;
          }

        case 'newPeer':
          {
            final Map<String, dynamic> newPeer =
                Map<String, dynamic>.from(notification['data']);
            peersBloc.add(PeerAdd(newPeer: newPeer));
            break;
          }

        case 'peerClosed':
          {
            String peerId = notification['data']['peerId'];
            peersBloc.add(PeerRemove(peerId: peerId));
            break;
          }

        case 'recordingStarted':
          {
            this.eventEmitter.emit('recording-started');
            break;
          }

        case 'recordingStopped':
          {
            this.eventEmitter.emit('recording-stopped');
            break;
          }

        case "livestreamStarted":
          {
            this.eventEmitter.emit('livestream-started');

            break;
          }

        case "livestreamStopped":
          {
            this.eventEmitter.emit('livestream-stopped');

            break;
          }

        case "activeSpeaker":
          {
            String? peerId = notification['data']['peerId'];

            if (peerId != roomBloc.state.activeSpeakerId) {
              roomBloc.add(RoomSetActiveSpeakerId(speakerId: peerId));
            }

            break;
          }
        case "pubsubMessage":
          {
            this.eventEmitter.emit("pubsub-message", notification['data']);
            break;
          }

        default:
          break;
      }
    };
  }

  Future requestEntry() async {
    await _webSocket!.socket.request("requestEntry", <String, dynamic>{
      'name': this.displayName,
    });
  }

  Future respondEntry(peerId, decision) async {
    await _webSocket!.socket.request(
        "respondEntry", <String, dynamic>{'id': peerId, 'decision': decision});
  }

  Future<Map<String, dynamic>> getDeviceInfo() async {
    String platform;

    if (Platform.isAndroid) {
      platform = "android";
    } else if (Platform.isIOS) {
      platform = "ios";
    } else {
      platform = "web";
    }

    final deviceInfo = LinkedHashMap<String, dynamic>();

    deviceInfo.addAll(<String, dynamic>{
      "sdkType": 'flutter',
      "sdkVersion": sdkVersion,
      "platform": platform
    });

    var deviceUserAgent = await Videosdk.deviceInfo;

    if (deviceUserAgent != null) {
      deviceInfo.addAll(<String, dynamic>{"deviceUserAgent": deviceUserAgent});
    }

    return deviceInfo;
  }

  // PubSub
  Future<void> pubsubPublish({topic, message, options}) async {
    try {
      await _webSocket!.socket.request('pubsubPublish', {
        'topic': topic,
        'message': message,
        'options': options,
      });
    } catch (error) {
      log("pubsubPublish() | failed: $error");
    }
  }

  Future<dynamic> pubsubSubscribe(topic) async {
    try {
      var msgList =
          await _webSocket!.socket.request('pubsubSubscribe', {'topic': topic});
      return msgList;
    } catch (error) {
      log("pubsubSubscribe() | failed: $error");
    }
  }

  Future<void> pubsubUnsubscribe(topic) async {
    try {
      await _webSocket!.socket.request('pubsubUnsubscribe', {'topic': topic});
    } catch (error) {
      log("pubsubUnsubscribe() | failed: $error");
    }
  }

  // Update usage
  _updateUsageStats() {
    if (_recvTransport?.consumers.isEmpty ?? true) return;

    List<Map<String, dynamic>> onUpdateUsageMap = [];

    _recvTransport?.consumers.forEach((consumerId, consumer) {
      if (consumer.kind == "video" && !consumer.paused) {
        var isScreenShare = consumer.appData['share'] ?? false;
        var peerData;

        var selectedResolution =
            getResolutionFromSpatialLayer(consumer.spatialLayer);

        num trackResolutionRounded =
            this._getRoundedResolution(selectedResolution);

        // check if this is peer's video stream or screen share
        if (isScreenShare) {
          peerData = "${consumer.appData['peerId']}-screen";
        } else {
          peerData = consumer.appData['peerId'];
        }

        //check if it's not exists then set this resolution
        if (!this._usageDataPointsPeerMaxRes.containsKey(peerData)) {
          this._usageDataPointsPeerMaxRes[peerData] = trackResolutionRounded;
        } else {
          //if exists then check if it's greater than the current max

          if (this._usageDataPointsPeerMaxRes[peerData] <
              trackResolutionRounded) {
            this._usageDataPointsPeerMaxRes[peerData] = trackResolutionRounded;
          }
        }
        onUpdateUsageMap.add(<String, dynamic>{
          "peerId": peerData,
          "resolution": trackResolutionRounded,
        });
      } else if (consumer.kind == "audio" && !consumer.paused) {
        this._audioConsumption++;
      }
    });
  }

  int getResolutionFromSpatialLayer(int spatialLayer) {
    switch (spatialLayer) {
      case -1:
        return 0;
      case 0:
        return 180;
      case 1:
        return 360;
      case 2:
        return 720;
      case 3:
        return 1080;
      default:
        return 360;
    }
  }

  num _getRoundedResolution(num? resolution) {
    if ((resolution == null) || (resolution == 0))
      return 0;
    else if (resolution <= 180)
      return 180;
    else if (resolution > 180 && resolution <= 360)
      return 360;
    else if (resolution > 360 && resolution <= 720)
      return 720;
    else if (resolution > 720 && resolution <= 1080)
      return 1080;
    else if (resolution > 1080 && resolution <= 1440)
      return 1440;
    else
      return 2160;
  }
}
